@@ -1,5 +1,3 @@
-# -*- coding: utf-8 -*-
-
 # Licensed under the Apache License, Version 2.0 (the "License"); you may
 # not use this file except in compliance with the License. You may obtain
 # a copy of the License at
@@ -14,19 +12,26 @@
 
 from collections import OrderedDict
 import os
+import re
 
 from docutils import nodes
+from docutils.parsers import rst
 from docutils.parsers.rst.directives.tables import Table
 from docutils.statemachine import ViewList
 import pbr.version
-from sphinx.util.compat import Directive
 from sphinx.util.osutil import copyfile
 import yaml
 
+from os_api_ref.http_codes import http_code
+from os_api_ref.http_codes import http_code_html
+from os_api_ref.http_codes import HTTPResponseCodeDirective
 
 __version__ = pbr.version.VersionInfo(
     'os_api_ref').version_string()
 
+# This is to allow for a graceful swap from oslosphinx to openstackdocstheme
+THEME = 'openstackdocstheme'
+
 
 """This provides a sphinx extension able to create the HTML needed
 for the api-ref website.
@@ -60,7 +65,8 @@ in both naming and ordering of parameters at every declaration.
 """
 
 
-def ordered_load(stream, Loader=yaml.Loader, object_pairs_hook=OrderedDict):
+def ordered_load(
+        stream, Loader=yaml.SafeLoader, object_pairs_hook=OrderedDict):
     """Load yaml as an ordered dict
 
     This allows us to inspect the order of the file on disk to make
@@ -117,7 +123,7 @@ class rest_expand_all(nodes.Part, nodes.Element):
     pass
 
 
-class RestExpandAllDirective(Directive):
+class RestExpandAllDirective(rst.Directive):
     # This tells sphinx that the directive will need to generate
     # content during the final build phase.
     has_content = True
@@ -127,12 +133,14 @@ class RestExpandAllDirective(Directive):
         node = rest_expand_all()
         max_ver = app.config.os_api_ref_max_microversion
         min_ver = app.config.os_api_ref_min_microversion
+        releases = app.config.os_api_ref_release_microversions
         node['major'] = None
         try:
             if max_ver.split('.')[0] == min_ver.split('.')[0]:
                 node['max_ver'] = int(max_ver.split('.')[1])
                 node['min_ver'] = int(min_ver.split('.')[1])
                 node['major'] = int(max_ver.split('.')[0])
+                node['releases'] = releases
         except ValueError:
             # TODO(sdague): warn that we're ignoring this all
             pass
@@ -141,7 +149,7 @@ class RestExpandAllDirective(Directive):
         return [node]
 
 
-class RestMethodDirective(Directive):
+class RestMethodDirective(rst.Directive):
 
     # this enables content in the directive
     has_content = True
@@ -171,6 +179,12 @@ class RestMethodDirective(Directive):
 
         node['method'] = method
         node['url'] = url
+
+        # Extract the path parameters from the url
+        env = self.state.document.settings.env
+        env.path_params = []
+        env.path_params = re.findall("{[a-zA-Z][a-zA-Z_0-9]*}", url)
+
         node['target'] = self.state.parent.attributes['ids'][0]
         node['css_classes'] = ""
         if node['min_version']:
@@ -203,20 +217,28 @@ class RestParametersDirective(Table):
         if fpath in YAML_CACHE:
             return YAML_CACHE[fpath]
 
+        lookup = {}
         # self.app.info("Fpath: %s" % fpath)
         try:
             with open(fpath, 'r') as stream:
                 lookup = ordered_load(stream)
-                self._check_yaml_sorting(fpath, lookup)
         except IOError:
-            self.env.warn(
-                self.env.docname,
-                "Parameters file %s not found" % fpath)
+            self.app.warn(
+                "Parameters file %s not found" % fpath,
+                (self.env.docname, None))
             return
         except yaml.YAMLError as exc:
             self.app.warn(exc)
             raise
 
+        if lookup:
+            self._check_yaml_sorting(fpath, lookup)
+        else:
+            self.app.warn(
+                "Parameters file is empty %s" % fpath,
+                (self.env.docname, None))
+            return
+
         YAML_CACHE[fpath] = lookup
         return lookup
 
@@ -272,29 +294,58 @@ class RestParametersDirective(Table):
         This allows use to reference an external file for the actual
         parameter definitions.
         """
+
         lookup = self._load_param_file(fpath)
+        if not lookup:
+            return
 
         content = "\n".join(self.content)
-        parsed = yaml.load(content)
+        parsed = yaml.safe_load(content)
         # self.app.info("Params loaded is %s" % parsed)
         # self.app.info("Lookup table looks like %s" % lookup)
         new_content = list()
         for paramlist in parsed:
+            if not isinstance(paramlist, dict):
+                self.app.warn(
+                    ("Invalid parameter definition ``%s``. Expected "
+                     "format: ``name: reference``. "
+                     " Skipping." % paramlist),
+                    (self.state_machine.node.source,
+                     self.state_machine.node.line))
+                continue
             for name, ref in paramlist.items():
                 if ref in lookup:
                     new_content.append((name, lookup[ref]))
                 else:
                     # TODO(sdague): this provides a kind of confusing
-                    # error message because env.warn isn't meant to be
+                    # error message because app.warn isn't meant to be
                     # used this way, however it does provide a way to
                     # track down where the parameters list is that is
                     # wrong. So it's good enough for now.
-                    self.env.warn(
-                        "%s:%s " % (
-                            self.state_machine.node.source,
-                            self.state_machine.node.line),
+                    self.app.warn(
                         ("No field definition for ``%s`` found in ``%s``. "
-                         " Skipping." % (ref, fpath)))
+                         " Skipping." % (ref, fpath)),
+                        (self.state_machine.node.source,
+                         self.state_machine.node.line))
+
+                # Check for path params in stanza
+                for i, param in enumerate(self.env.path_params):
+                    if (param.rstrip('}').lstrip('{')) == name:
+                        del self.env.path_params[i]
+                        break
+                    else:
+                        continue
+
+        if len(self.env.path_params) is not 0:
+            # Warn that path parameters are not set in rest_parameter
+            # stanza and will not appear in the generated table.
+            for param in self.env.path_params:
+                self.app.warn(
+                    ("No path parameter ``%s`` found in rest_parameter"
+                     " stanza.\n"
+                     % param.rstrip('}').lstrip('{')),
+                    (self.state_machine.node.source,
+                     self.state_machine.node.line))
 
         # self.app.info("New content %s" % new_content)
         self.yaml = new_content
@@ -329,8 +380,13 @@ class RestParametersDirective(Table):
         # TODO(sdague): it would be good to dynamically set column
         # widths (or basically make the colwidth thing go away
         # entirely)
-        self.options['widths'] = (20, 10, 10, 60)
+        self.options['widths'] = [20, 10, 10, 60]
         self.col_widths = self.get_column_widths(self.max_cols)
+        if isinstance(self.col_widths, tuple):
+            # In docutils 0.13.1, get_column_widths returns a (widths,
+            # colwidths) tuple, where widths is a string (i.e. 'auto').
+            # See https://sourceforge.net/p/docutils/patches/120/.
+            self.col_widths = self.col_widths[1]
         # Actually convert the yaml
         title, messages = self.make_title()
         # self.app.info("Title %s, messages %s" % (title, messages))
@@ -385,7 +441,7 @@ class RestParametersDirective(Table):
                                         str(min_version).replace('.', '_'))
                     classes.append(min_ver_css_name)
                 if max_version:
-                    desc += ("\n\n**Deprecated in version %s**\n" %
+                    desc += ("\n\n**Available until version %s**\n" %
                              max_version)
                     max_ver_css_name = ("rp_max_ver_" +
                                         str(max_version).replace('.', '_'))
@@ -441,14 +497,25 @@ class RestParametersDirective(Table):
 
 def rest_method_html(self, node):
     tmpl = """
-<div class="row operation-grp %(css_classes)s">
-    <div class="col-md-1 operation">
-    <a name="%(target)s" class="operation-anchor" href="#%(target)s">
+<div class="operation-grp %(css_classes)s">
+<div class="row">
+    <div class="col-md-2">
+    <div class="operation">
+    <a name="%(target)s" class="operation-anchor" href="#%(target)s"
+      onclick="window.location.hash = hash;"
+      >
       <span class="glyphicon glyphicon-link"></span></a>
     <span class="label label-%(method)s">%(method)s</span>
     </div>
-    <div class="col-md-5">%(url)s</div>
-    <div class="col-md-5">%(desc)s</div>
+    </div>
+    <div class="col-md-9">
+    <div class="row">
+        <div class="endpoint-container">
+        <div class="row col-md-12">%(url)s</div>
+        <div class="row col-md-12"><p class="url-subtitle">%(desc)s</p></div>
+        </div>
+    </div>
+    </div>
     <div class="col-md-1">
     <button
        class="btn btn-info btn-sm btn-detail"
@@ -457,6 +524,7 @@ def rest_method_html(self, node):
        id="%(target)s-detail-btn"
        >detail</button>
     </div>
+</div>
 </div>"""
 
     node['url'] = node['url'].replace(
@@ -474,7 +542,9 @@ def rest_expand_all_html(self, node):
     tmpl = """
 <div class="row">
 %(extra_js)s
-<div class=col-md-11>%(selector)s</div>
+<div class="col-md-2 col-md-offset-9">
+%(selector)s
+</div>
 <div class=col-md-1>
     <button id="expand-all"
        data-toggle="collapse"
@@ -483,28 +553,64 @@ def rest_expand_all_html(self, node):
 </div>
 </div>"""
 
+    node.setdefault('selector', "")
+    node.setdefault('extra_js', "")
+
     if node['major']:
-        selector = """
-<div class="btn-group" role="group" aria-label="...">
-    <button type="button"
-        class="btn btn-default mv_selector active">All</button>\n"""
-        for x in range(node['min_ver'], node['max_ver'] + 1):
-            selector += ('<button type="button" class="btn btn-default '
-                         'mv_selector">%d.%d</button>\n' % (node['major'], x))
-        selector += "</div>"
-        node['selector'] = selector
-        node['extra_js'] = ("<script>os_max_mv = %d; "
-                            "os_min_mv = %d;</script>") % (
-                                node['max_ver'],
-                                node['min_ver'])
-    else:
-        node['selector'] = ""
-        node['extra_js'] = ""
+        node['selector'], node['extra_js'] = create_mv_selector(node)
 
     self.body.append(tmpl % node)
     raise nodes.SkipNode
 
 
+def create_mv_selector(node):
+
+    mv_list = '<option value="" selected="selected">All</option>'
+
+    for x in range(node['min_ver'], node['max_ver'] + 1):
+        mv_list += build_mv_item(node['major'], x, node['releases'])
+
+    selector_tmpl = """
+<form class=form-inline">
+<div class="form-group">
+<label class="control-label">
+Microversions
+</label>
+<select class="combobox form-control" id="mv_select">
+    %(mv_list)s
+</select>
+</div>
+</form>
+"""
+
+    js_tmpl = """
+<script>
+    os_max_mv = %(max)d;
+    os_min_mv = %(min)d;
+</script>
+"""
+
+    selector_content = {
+        'mv_list': mv_list
+    }
+
+    js_content = {
+        'min': node['min_ver'],
+        'max': node['max_ver']
+    }
+
+    return selector_tmpl % selector_content, js_tmpl % js_content
+
+
+def build_mv_item(major, micro, releases):
+    version = "%d.%d" % (major, micro)
+    if version in releases:
+        return '<option value="%s">%s - %s</option>' % (
+            version, version, releases[version].capitalize())
+    else:
+        return '<option value="%s">%s</option>' % (version, version)
+
+
 def resolve_rest_references(app, doctree):
     for node in doctree.traverse():
         if isinstance(node, rest_method):
@@ -538,30 +644,36 @@ def resolve_rest_references(app, doctree):
 
 
 def copy_assets(app, exception):
-    assets = ('bootstrap.min.css', 'api-site.css',
-              'bootstrap.min.js', 'api-site.js',
-              'glyphicons-halflings-regular.ttf',
-              'glyphicons-halflings-regular.woff')
+    assets = ('api-site.css', 'api-site.js', 'combobox.js')
+    fonts = (
+        'glyphicons-halflings-regular.ttf',
+        'glyphicons-halflings-regular.woff'
+    )
     if app.builder.name != 'html' or exception:
         return
     app.info('Copying assets: %s' % ', '.join(assets))
+    app.info('Copying fonts: %s' % ', '.join(fonts))
     for asset in assets:
         dest = os.path.join(app.builder.outdir, '_static', asset)
         source = os.path.abspath(os.path.dirname(__file__))
         copyfile(os.path.join(source, 'assets', asset), dest)
+    for font in fonts:
+        dest = os.path.join(app.builder.outdir, '_static/fonts', font)
+        source = os.path.abspath(os.path.dirname(__file__))
+        copyfile(os.path.join(source, 'assets', font), dest)
 
 
 def add_assets(app):
-    app.add_stylesheet('bootstrap.min.css')
     app.add_stylesheet('api-site.css')
-    app.add_javascript('bootstrap.min.js')
     app.add_javascript('api-site.js')
+    app.add_javascript('combobox.js')
 
 
 def setup(app):
     # Add some config options around microversions
     app.add_config_value('os_api_ref_max_microversion', '', 'env')
     app.add_config_value('os_api_ref_min_microversion', '', 'env')
+    app.add_config_value('os_api_ref_release_microversions', '', 'env')
     # TODO(sdague): if someone wants to support latex/pdf, or man page
     # generation using these stanzas, here is where you'd need to
     # specify content specific renderers.
@@ -569,11 +681,14 @@ def setup(app):
                  html=(rest_method_html, None))
     app.add_node(rest_expand_all,
                  html=(rest_expand_all_html, None))
+    app.add_node(http_code,
+                 html=(http_code_html, None))
 
     # This specifies all our directives that we're adding
     app.add_directive('rest_parameters', RestParametersDirective)
     app.add_directive('rest_method', RestMethodDirective)
     app.add_directive('rest_expand_all', RestExpandAllDirective)
+    app.add_directive('rest_status_code', HTTPResponseCodeDirective)
 
     # The doctree-read hook is used do the slightly crazy doc
     # transformation that we do to get the rest_method document
